/*
    snd2mid - simple portable tool for convert sound files to midi format.
    Copyright (C) 2005  Alexey V. Novikov <shader@yandex.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>

#include <sndfile.h>
#include <fftw3.h>

#include "midifile/midifile.h"
/*
    Function declarations from midifile.c
*/
extern void mf_write_tempo(unsigned long tempo);
extern int mf_write_midi_event(unsigned long delta_time,
                               unsigned int type,
                               unsigned int chan,
                               unsigned char *data,
                               unsigned long size);

#define PROGNAME "snd2mid"

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))
#define NOTES 127
char note_name[12][3] = {"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "H"};

#define DEFAULT_TRACK_SIZE 512
#define DEFAULT_THRESHOLD 0.2
float threshold = DEFAULT_THRESHOLD;

/* 
    Variable for input sound file
	sfinfo		- input sound file info structure
	inbuf		- buffer for storing data that read from input
	inbuf_time	- time of fragment stored in inbuf
	inbuf_size	- size of inbuf
*/
SF_INFO sfinfo;
int inbuf_size;
float *inbuf;
#define DEFAULT_INBUF_TIME 0.05
float inbuf_time = DEFAULT_INBUF_TIME;

/*
    Hamming window
*/
float *hamming;

/*
    Variables for FFTW
	fftw_size	- size of FFTW array
	fftw_in,
	fftw_out	- in and out arrays for FFTW
	fftw_p		- FFTW plan
*/
int fftw_size;
fftwf_complex *fftw_in, *fftw_out;
fftwf_plan fftw_p;

/*
    Type and variables for storing result
	note_t		- type for storing result
	    freq	- note frequency
	    name	- symbolic name of note
	    octave	- note's octave
	    ind		- index of note in fftw_out array
	    track	- pointer on channel-based arrays of values
	    note_number - note number in MIDI standard
*/
typedef struct{
    int freq;
    char name[3];
    int octave;
    int ind;
    float **track;
    int *track_size;
    int note_number;
} note_t;
note_t notes[NOTES];

/*
    Maximum of note's values for given sample
*/
float **max_value;

/*
    Sample counter
*/
int sample;

/*
    MIDI file pointer
*/
FILE *fp;

/*
    Variable controlled by command line options
*/
int debug = 0;

/*
    Key filter
*/
int kf_filter[NOTES];
int kf_maj[] = { 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1};
int kf_min[] = { 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 1, 1};

/*
    Callback functions for mfwrite
*/
int midiputc(c)
{
    return putc(c, fp);
}

int midiwritechannel(int channel)
{
    int i = 0;
    char data[2];
    mf_write_tempo((long)500000); /* 120 beats/per/second */

    int delta = 0;
    float maxi = 0.0, maxi_old = 0.0;
    while(i <= sample){
        int note;
        maxi = max((max_value[channel][i] + maxi) / 2.0, max_value[channel][i]);
	/* first "off" all previously "on" notes */
        for(note = 0; note < NOTES; note++){
            if(i == 0) break;
            if((i == sample || notes[note].track[channel][i] <= maxi * threshold) &&
               (notes[note].track[channel][i-1] > maxi_old * threshold)){
                data[0] = notes[note].note_number;
                data[1] = 0;
                mf_write_midi_event(delta, note_off, channel, data, 2);
                delta = 0;
            }
        }
        for(note = 0; note < NOTES; note++){
            if(i == sample || maxi == 0.0) break;
            if((notes[note].track[channel][i] > maxi * threshold) &&
               (i == 0 || notes[note].track[channel][i-1] <= maxi_old * threshold)){
                data[0] = notes[note].note_number;
                data[1] = (int)(notes[note].track[channel][i]*100.0/maxi);
                mf_write_midi_event(delta, note_on, channel, data, 2);
                delta = 0;
            }
        }
        delta += (int)(inbuf_time*1000.0);
        i++;
	maxi_old = maxi;
    }
    return 1;
}

/*
    Write out result MIDI file
*/
void write_midifile(const char* fname)
{
    if((fp = fopen(fname, "wb")) == NULL){
	fprintf(stderr, "Unable to open file %s for writing.\n", fname);
	return;
    }

    Mf_putc = midiputc;
    Mf_writetrack = midiwritechannel;

    /* write a single track */
    mfwrite(0, sfinfo.channels, 480, fp);
    fclose(fp);
}

/*
    Filter note value based on key
*/
void filter_notes_kf(int channel)
{
    int note, i;
    for(i = 0; i < sample; i++){
	max_value[channel][i] = 0.0;
        for(note = 0; note < NOTES; note++){
            if(kf_filter[note] == 0)
                notes[note].track[channel][i] = 0.0;
	    max_value[channel][i] = max(max_value[channel][i], notes[note].track[channel][i]);
        }
    }
}

/*
    Temporal filter note values
*/
void filter_notes_temp(int channel)
{
    int note, i;
    for(i = 0; i < sample; i++){
	max_value[channel][i] = 0.0;
        for(note = 0; note < NOTES; note++){
            float xvalue = 1.0*notes[note].track[channel][i];
            float xcount = 1.0;
            if(i > 0){
                xvalue += 1.0*notes[note].track[channel][i - 1];
                xcount += 1.0;
            }
	    if(i > 1){
                xvalue += 1.0*notes[note].track[channel][i - 2];
                xcount += 1.0;
	    }
            if(i + 2 < sample){
                xvalue += 1.0*notes[note].track[channel][i + 2];
                xcount += 1.0;
            }
            if(i + 1 < sample){
                xvalue += 1.0*notes[note].track[channel][i + 1];
                xcount += 1.0;
            }
            notes[note].track[channel][i] = xvalue/xcount;
	    max_value[channel][i] = max(max_value[channel][i], notes[note].track[channel][i]);
        }
    }
}

/*
    Set note values for single sample
*/
void set_note_values(int channel)
{
    int note, walk_down = 0;
    float old_value = -1.0, new_value;
    if(notes[0].track_size[channel] <= sample){
	max_value[channel] = realloc(max_value[channel],
	                             (notes[0].track_size[channel] + DEFAULT_TRACK_SIZE) * sizeof(float));
	if(!max_value[channel]){
	    fprintf(stderr, "max_value[%d]: Can't allocate %d bytes\n",
		    channel, (notes[0].track_size[channel] + DEFAULT_TRACK_SIZE) * sizeof(float));
	    exit(1);
	}
    }
    max_value[channel][sample] = 0.0;
    /* find spectrum's peeks */
    for(note = 0; note < NOTES; note++){
        if(notes[note].track_size[channel] <= sample){
            notes[note].track[channel] = realloc(notes[note].track[channel],
                                              (notes[note].track_size[channel] + DEFAULT_TRACK_SIZE) * sizeof(float));
	    if(!notes[note].track[channel]){
		fprintf(stderr, "notes[%d].track[%d]: Can't allocate %d bytes\n",
			note, channel, (notes[note].track_size[channel] + DEFAULT_TRACK_SIZE) * sizeof(float));
		exit(1);
	    }
            notes[note].track_size[channel] += DEFAULT_TRACK_SIZE;
        }
        notes[note].track[channel][sample] = 0.0;
        new_value = notes[note].ind < fftw_size ? sqrt(fftw_out[notes[note].ind][0] * fftw_out[notes[note].ind][0] + fftw_out[notes[note].ind][1] * fftw_out[notes[note].ind][1]) : 0.0;
        if(new_value < old_value && !walk_down){
            notes[note - 1].track[channel][sample] = old_value;
            walk_down = 1;
        }
        else if(new_value > old_value)
            walk_down = 0;
	max_value[channel][sample] = max(max_value[channel][sample], new_value);
        old_value = new_value;
    }
}

/*
    Initialize "notes" and "max_value" arrays
*/
void init_notes()
{
    int octave = 0, note, channel;
    float delta = pow(2.0, 1.0 / 12.0);
    float x = 440.0 / 64.0 * delta * delta * delta;
    for(note = 0; note < NOTES; note++){
        notes[note].track_size = malloc(sfinfo.channels * sizeof(int));
	if(!notes[note].track_size){
    	    fprintf(stderr, "notes[%d].track_size: Can't allocate %d bytes\n",
        	    note, sfinfo.channels * sizeof(int));
    	    exit(1);
	}
        notes[note].track = malloc(sfinfo.channels * sizeof(float*));
	if(!notes[note].track){
    	    fprintf(stderr, "notes[%d].track: Can't allocate %d bytes\n",
        	    note, sfinfo.channels * sizeof(float*));
    	    exit(1);
	}
        for(channel = 0; channel < sfinfo.channels; channel++){
            notes[note].track[channel] = malloc(DEFAULT_TRACK_SIZE * sizeof(float));
	    if(!notes[note].track[channel]){
    		fprintf(stderr, "notes[%d].track[%d]: Can't allocate %d bytes\n",
        	        note, channel, DEFAULT_TRACK_SIZE * sizeof(float));
	        exit(1);
	    }
            notes[note].track_size[channel] = DEFAULT_TRACK_SIZE;
        }
        notes[note].octave = octave;
        strcpy(notes[note].name, note_name[note % 12]);
        notes[note].freq = x;
        notes[note].ind = (int)(x * (float)fftw_size / (float)sfinfo.samplerate);
        notes[note].note_number = note;
        x *= delta;
	octave += note % 12 == 11 ? 1 : 0;
    }
    max_value = malloc(sfinfo.channels * sizeof(float*));
    if(!max_value){
        fprintf(stderr, "max_value: Can't allocate %d bytes\n",
		sfinfo.channels * sizeof(float*));
    	exit(1);
    }
    for(channel = 0; channel < sfinfo.channels; channel++){
        max_value[channel] = malloc(DEFAULT_TRACK_SIZE * sizeof(float));
	if(!max_value[channel]){
    	    fprintf(stderr, "max_value[%d]: Can't allocate %d bytes\n",
		    channel, DEFAULT_TRACK_SIZE * sizeof(float));
    	    exit(1);
	}
    }
}

/*
    Destroy "notes" and "max_value" arrays
*/
void destroy_notes()
{
    int note, channel;
    for(note = 0; note < NOTES; note++){
        for(channel = 0; channel < sfinfo.channels; channel++)
            free(notes[note].track[channel]);
        free(notes[note].track);
        free(notes[note].track_size);
    }
    for(channel = 0; channel < sfinfo.channels; channel++)
        free(max_value[channel]);
    free(max_value);
}

/*
    Initialize FFTW
*/
void init_fftw()
{
    fftw_size = (int)pow(2.0, (int)(log((float)sfinfo.samplerate)/log(2.0)));
    if(debug) printf("fftw_size=%d\n", fftw_size);
    fftw_in = fftwf_malloc(sizeof(fftw_complex) * fftw_size);
    if(!fftw_in){
        fprintf(stderr, "fftw_in: Can't allocate %d bytes\n",
            fftw_size * sizeof(fftw_complex));
        exit(1);
    }
    fftw_out = fftwf_malloc(sizeof(fftw_complex) * fftw_size);
    if(!fftw_out){
        fprintf(stderr, "fftw_out: Can't allocate %d bytes\n",
            fftw_size * sizeof(fftw_complex));
        exit(1);
    }
    fftw_p = fftwf_plan_dft_1d(fftw_size, fftw_in, fftw_out, FFTW_FORWARD, FFTW_ESTIMATE);
}

/*
    Free memory allocated for FFTW
*/
void destroy_fftw()
{
    fftwf_destroy_plan(fftw_p);
    fftwf_free(fftw_in);
    fftwf_free(fftw_out);
}

/*
    Prepare Hamming window
*/
void init_hamming()
{
    int size = inbuf_size / sfinfo.channels;
    hamming = malloc(size * sizeof(float));
    if(!hamming){
        fprintf(stderr, "hamming: Can't allocate %d bytes\n",
            size * sizeof(float));
        exit(1);
    }
    int i;
    for(i = 0; i < size; i++)
        hamming[i] = 0.54 - 0.46 * cos(2.0 * M_PI * (float)i / (float)(size - 1));
}

/*
    Initialize inbuf
*/
void init_inbuf()
{
    inbuf_size = sfinfo.channels * max((int)((float)sfinfo.samplerate * inbuf_time), 10.0);
    inbuf = malloc(inbuf_size * sizeof(float));
    if(!inbuf){
        fprintf(stderr, "inbuf: Can't allocate %d bytes\n", inbuf_size * sizeof(float));
        exit(1);
    }
    if(debug) printf("inbuf_size=%d\n", inbuf_size);
}

/*
    Main loop
*/
void main_loop(SNDFILE *infile)
{
    int inbuf_len = 0;
    sample = 0;
    while(1){
        inbuf_len = sf_read_float(infile, inbuf, inbuf_size);
        if(inbuf_len <= 0) break;

        int ch, channel_len = inbuf_len / sfinfo.channels;
        for(ch = 0; ch < sfinfo.channels; ch++){
            /* initialize fftw_in */
	    int i;
            for(i = 0; i < fftw_size; i++){
                fftw_in[i][0] = i < channel_len ? hamming[i] * inbuf[i * sfinfo.channels + ch] : 0.0;
                fftw_in[i][1] = 0.0;
            }
            /* execute plan */
            fftwf_execute(fftw_p);

            /* setup values for notes */
            set_note_values(ch);
        }
        sample++;
    }
}

/*
    Printout usage
*/
void usage()
{
    printf("Usage: %s [options] <audiofile> <midifile>\n", PROGNAME);
    printf("Options:\n");
    printf("  -v, --verbose        Debug output.\n");
    printf("  -s n, --slice=n      specify slice time in milliseconds (default %d)\n", (int)(DEFAULT_INBUF_TIME * 1000.0));
    printf("  -t n, --threshold=n  specify threshold from 0.01 to 1 (default %.2f)\n", DEFAULT_THRESHOLD);
    printf("  -k n, --key=n        specify key filter (e.g. C-maj or F#-min)\n");
    printf("  -h, --help           Print this message and exit.\n");
}

/*
    Main function
*/
int main(int argc, char *argv[])
{
    int option_index = 0;
    int i;
    static struct option long_options[] = {
	{"verbose", 0, 0, 'v'},
	{"slice", 1, 0, 's'},
	{"threshold", 1, 0, 't'},
	{"key", 1, 0, 'k'},
	{"help", 0, 0, 'h'},
	{0, 0, 0, 0}
    };

    /* init kf_filter array - default not filter anything */
    for(i = 0; i < NOTES; i++)
        kf_filter[i] = 1;
    
    int c;
    while ((c = getopt_long(argc, argv, "hvs:t:k:", long_options, &option_index)) != -1) {
	switch (c) {
	    case 'v':
		debug = 1;
		break;
	    case 's':
        	inbuf_time = 0.001 * (float)atoi(optarg);
		if(inbuf_time <= 0.0 || inbuf_time >= 0.5){
		    usage();
		    return 1;
		}
        	break;
	    case 't':
        	threshold = atof(optarg);
		if(threshold < 0.01 || threshold > 1.0){
		    usage();
		    return 1;
		}
        	break;
	    case 'k':{
	        char *p = strchr(optarg, '-');
	        if(p - optarg == 1 || p - optarg == 2){
	            char str[3];
	            strncpy(str, optarg, p - optarg);
	            str[p - optarg] = '\0';
	            int note_num = -1;
	            for(i = 0; i < 12; i++)
	                if(!strcmp(str, note_name[i])){
	                    note_num = i;
	                    break;
	                }
	            if(note_num < 0){
		        usage();
		        return 1;
	            }
	            if(!strcmp(p + 1, "maj")){
	                for(i = 0; i < NOTES; i++)
	                    kf_filter[i] = kf_maj[(i - note_num) % 12];
	            }
	            else if(!strcmp(p + 1, "min")){
	                for(i = 0; i < NOTES; i++)
	                    kf_filter[i] = kf_min[(i - note_num) % 12];
	            }
	            else{
		        usage();
		        return 1;
	            }
        	}
		else{
		    usage();
		    return 1;
		}
        	break;}
	    case 'h':
		usage();
		return 0;
	    default:
		usage();
		return 1;
	}
    }

    argc -= optind;
    argv = &argv[optind];

    if (argc != 2){
    	usage();
	exit(1);
    }

    /* open soundfile */
    sfinfo.format = 0;
    SNDFILE *sndfile = sf_open(argv[0], SFM_READ, &sfinfo);
    if(!sndfile){
        fprintf(stderr, "Error while opening file %s: %s\n",
            argv[0], sf_error_number(sf_error(sndfile)));
        exit(1);
    }
    if(debug)
	printf("samplerate=%d, channels=%d, format=0x%X\n",
    	    sfinfo.samplerate, sfinfo.channels, sfinfo.format);

    /* prepare */
    init_fftw();
    init_notes();
    init_inbuf();
    init_hamming();

    /* main work */
    main_loop(sndfile);

    /* filter notes */
    int channel;
    for(channel = 0; channel < sfinfo.channels; channel++){
        filter_notes_kf(channel);
        filter_notes_temp(channel);
    }
    /* write result */
    write_midifile(argv[1]);

    /* finalization */
    destroy_fftw();
    destroy_notes();
    free(hamming);
    free(inbuf);
    sf_close(sndfile);
    return 0;
}
